package com.cesys.utils;

import com.sun.jna.Native;
import com.sun.jna.Platform;

import java.io.*;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;

public class Loader {
    private enum Architecture {
        Win64("windows", ".dll", ""),
        Linux64("linux-intel", ".so", "lib"),
        LinuxAarch64("linux-aarch64", ".so", "lib"),
        Darwin("darwin", ".dylib", "lib"),
        Unknown("", "", "")
        ;

        private final String _tag, _ext, _prefix;

        public String getTag() { return _tag; }
        public String getExt() { return _ext; }
        public String getPrefix() { return _prefix; }

        Architecture(String tag, String ext, String prefix) {
            _tag = tag;
            _ext = ext;
            _prefix = prefix;
        }
    }

    private static Architecture getArchitecture() {
        if (Platform.isWindows() && Platform.is64Bit())
            return Architecture.Win64;

        if (Platform.isLinux() && Platform.is64Bit()) {
            if (Platform.isARM())
                return Architecture.LinuxAarch64;
            else
                return Architecture.Linux64;
        }

        if (Platform.isMac())
            return Architecture.Darwin;

        return Architecture.Unknown;
    }

    public static String prepareLib(String pathStub, String resourceStub, String nameStub) {
        final Architecture arch = getArchitecture();

        final String tmpPath = Paths.get(System.getProperty("user.home"), pathStub).toAbsolutePath().toString();
        final File path = new File(tmpPath);
        if (!path.exists()) {
            if (!path.mkdirs())
                throw new RuntimeException("Can't create temporary library folder for JNA");
        }

        final String outName = arch.getPrefix() + nameStub + arch.getExt();
        final Path p = Paths.get(tmpPath, outName).toAbsolutePath();
        if (p.toFile().exists()) {
            if (!p.toFile().delete()) {
                throw new RuntimeException(String.format("Can't delete existing lib '%s'", p.toString()));
            }
        }

        final String src = resourceStub + "/" + arch.getTag() + "/binary";
        try (final InputStream is = ClassLoader.getSystemResourceAsStream(src)) {
            if (is == null)
                throw new RuntimeException(String.format("Can't find resource '%s", src));
            Files.copy(is, p);
        } catch (IOException e) {
            e.printStackTrace();
        }

        System.setProperty("jna.library.path", tmpPath);

        Native.register(nameStub);

        return nameStub;
    }

    private static native void __wrapper_error(int index, byte[] buffer, int bufferSize, int[] kind);

    /**
     * Check return code of library calls. If an error is reported, an Exception with the library error text is thrown.
     * @param errorCode Code reported by the C-API.
     * @throws IllegalArgumentException If the underlying API detects an illegal argument.
     * @throws IndexOutOfBoundsException If the underlying API was forced to access an invalid array element.
     * @throws IllegalStateException If a method is called that should not be called in the current state.
     * @throws IOException Unrecoverable error in the lower API levels, in most cases an unpredictable communication error with the hardware.
     * @throws RuntimeException If the API reports an unknown error, this should never be occur except for the case of a bug inside the API.
     */
    public static void checkError(int errorCode) throws IOException {
        if (errorCode != 0) {
            byte[] buffer = new byte[1024];
            int[] kind = new int[1];

            __wrapper_error(errorCode, buffer, 1024, kind);

            String error = new String(buffer);

            switch (kind[0]) {
                case 0x80: throw new IllegalArgumentException(error);
                case 0x40: throw new IndexOutOfBoundsException(error);
                case 0x20: throw new IllegalStateException(error);
                case 0x10: throw new IOException(error);
            }

            // Throw exception with string.
            throw new RuntimeException("(internal) Unhandled error: " + error);
        }
    }

    /**
     * As Java has no uint32, the parameter must be a long, but it must result
     * in an int, that holds the binary version of the uint32. Additional,
     * an range check is done.
     * @param value
     * @return
     */
    public static int toUint32(long value) {
        final long maxValue = 4294967295L;
        final long minValue = 0;
        if (value < minValue || value > maxValue)
            throw new IllegalArgumentException("Value outside range of 32 bit unsigned int.");
        return (int)value;
    }

    /**
     * Convert int to 32 bit unsigned representation in a long.
     * @param value
     * @return
     */
    public static long fromUInt32(int value) {
        if (value >= 0)
            return (long)value;
        return 0x100000000L + (long)value;
    }
}
